@@ -13,107 +13,72 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-from neutron_lib import constants as n_const
-from neutron_lib import exceptions as n_exc
-from neutron_lib.plugins import directory
-
-from midonet.neutron._i18n import _, _LE, _LW
-from midonet.neutron.common import utils as c_utils
+from midonet.neutron.db import agent_membership_db as am_db
 from midonet.neutron.db import port_binding_db as pb_db
 from midonet.neutron.db import provider_network_db as pnet_db
-from midonet.neutron.midonet_v2 import managers
-from midonet.neutron.ml2 import sg_callback
 from midonet.neutron import plugin
-from midonet.neutron.services.qos import driver as qos_driver
-
-from neutron.api.v2 import attributes
+from neutron.common import constants as n_const
+from neutron.common import exceptions as n_exc
 from neutron.db import allowedaddresspairs_db as addr_pair_db
-from neutron.db import api as db_api
-from neutron.db import db_base_plugin_v2
-from neutron.db import portsecurity_db as ps_db
+from neutron.db import extraroute_db
 from neutron.extensions import allowedaddresspairs as addr_pair
 from neutron.extensions import extra_dhcp_opt as edo_ext
-from neutron.extensions import portsecurity as psec
 from neutron.extensions import providernet as pnet
 from neutron.extensions import securitygroup as ext_sg
+from neutron import i18n
+from oslo_db import api as oslo_db_api
 from oslo_db import exception as oslo_db_exc
-from oslo_log import helpers as log_helpers
 from oslo_log import log as logging
 from oslo_utils import excutils
 
 
 LOG = logging.getLogger(__name__)
+_LE = i18n._LE
+_LW = i18n._LW
 
 
 class MidonetPluginV2(plugin.MidonetMixinBase,
                       addr_pair_db.AllowedAddressPairsMixin,
+                      am_db.AgentMembershipDbMixin,
+                      extraroute_db.ExtraRoute_db_mixin,
                       pnet_db.MidonetProviderNetworkMixin,
-                      pb_db.MidonetPortBindingMixin,
-                      ps_db.PortSecurityDbMixin):
+                      pb_db.MidonetPortBindingMixin):
 
-    # REVISIT(yamamoto): Consider dropping agent and dhcp_agent_scheduler,
-    # probably after a cycle. (for M release)
-    # This plugin is for MidoNet v5.0 and later, where we have an
-    # alternative metadata proxy which doesn't require Neutron agents.
-    # NOTE(yamamoto): While the order in this list doesn't matter
-    # functionality-wise, it's alphabetically sorted for easier comparison.
-    _base_supported_extension_aliases = [
+    supported_extension_aliases = [
         'agent',
+        'agent-membership',
         'allowed-address-pairs',
         'binding',
         'dhcp_agent_scheduler',
         'external-net',
         'extra_dhcp_opt',
-        'port-security',
+        'extraroute',
         'provider',
         'quotas',
-        'security-group',
-        'subnet_allocation',
+        'router',
+        'security-group'
     ]
 
-    @property
-    def supported_extension_aliases(self):
-        if not hasattr(self, '_aliases'):
-            aliases = self._base_supported_extension_aliases
-            aliases += self.extension_manager.extension_aliases()
-            self._aliases = aliases
-        return self._aliases
-
     __native_bulk_support = True
-    __native_pagination_support = True
-    __native_sorting_support = True
 
     def __init__(self):
-        self.extension_manager = managers.ExtensionManager()
-        self.extension_manager.initialize()
         super(MidonetPluginV2, self).__init__()
         self.client.initialize()
-        qos_driver.register()
-        self.sec_handler = sg_callback.MidonetSecurityGroupsHandler(
-            self.client)
 
-    @db_api.retry_if_session_inactive()
     def create_network(self, context, network):
         LOG.debug('MidonetPluginV2.create_network called: network=%r', network)
 
         net_data = network['network']
-        tenant_id = net_data['tenant_id']
+        tenant_id = self._get_tenant_id_for_create(context, net_data)
         net_data['tenant_id'] = tenant_id
         self._ensure_default_security_group(context, tenant_id)
 
         with context.session.begin(subtransactions=True):
-            net_db = self.create_network_db(context, network)
-            net = self._make_network_dict(net_db, process_extensions=False,
-                                          context=context)
-            self.extension_manager.process_create_network(context, net_data,
-                                                          net)
+            net = super(MidonetPluginV2, self).create_network(context, network)
             net_data['id'] = net['id']
             self._process_l3_create(context, net, net_data)
             self._create_provider_network(context, net_data)
             self._extend_provider_network_dict(context, net)
-            if psec.PORTSECURITY in net_data:
-                self._process_network_port_security_create(context, net_data,
-                                                           net)
             self.client.create_network_precommit(context, net)
 
         try:
@@ -129,8 +94,6 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
                     LOG.exception(_LE("Failed to delete network %s"),
                                   net['id'])
 
-        self._apply_dict_extend_functions('networks', net, net_db)
-
         LOG.debug("MidonetPluginV2.create_network exiting: net=%r", net)
         return net
 
@@ -161,7 +124,6 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
 
         return [self._fields(net, fields) for net in nets]
 
-    @db_api.retry_if_session_inactive()
     def update_network(self, context, id, network):
         LOG.debug("MidonetPluginV2.update_network called: id=%(id)r, "
                   "network=%(network)r", {'id': id, 'network': network})
@@ -173,22 +135,8 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         with context.session.begin(subtransactions=True):
             net = super(MidonetPluginV2, self).update_network(
                 context, id, network)
-            self.extension_manager.process_update_network(context, net_data,
-                                                          net)
             self._process_l3_update(context, net, network['network'])
             self._extend_provider_network_dict(context, net)
-            if psec.PORTSECURITY in network['network']:
-                self._process_network_port_security_update(context,
-                                                           network['network'],
-                                                           net)
-            # NOTE(yamamoto): Retrieve the db object to get the correct
-            # revision
-            # ToDO(QoS): This would change once EngineFacade moves out
-            db_network = self._get_network(context, id)
-            # Expire the db_network in current transaction, so that the join
-            # relationship can be updated.
-            context.session.expire(db_network)
-            net = self._make_network_dict(db_network, context=context)
             self.client.update_network_precommit(context, id, net)
 
         self.client.update_network_postcommit(id, net)
@@ -196,12 +144,13 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         LOG.debug("MidonetPluginV2.update_network exiting: net=%r", net)
         return net
 
-    @db_api.retry_if_session_inactive()
+    @oslo_db_api.wrap_db_retry(max_retries=3, retry_interval=1,
+                               retry_on_request=True,
+                               retry_on_deadlock=True)
     def delete_network(self, context, id):
         LOG.debug("MidonetPluginV2.delete_network called: id=%r", id)
 
         with context.session.begin(subtransactions=True):
-            c_utils.check_delete_network_precommit(context, id)
             self._process_l3_delete(context, id)
             try:
                 super(MidonetPluginV2, self).delete_network(context, id)
@@ -219,14 +168,11 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
 
         LOG.debug("MidonetPluginV2.delete_network exiting: id=%r", id)
 
-    @db_api.retry_if_session_inactive()
     def create_subnet(self, context, subnet):
         LOG.debug("MidonetPluginV2.create_subnet called: subnet=%r", subnet)
 
         with context.session.begin(subtransactions=True):
             s = super(MidonetPluginV2, self).create_subnet(context, subnet)
-            self.extension_manager.process_create_subnet(context,
-                subnet['subnet'], s)
             self.client.create_subnet_precommit(context, s)
 
         try:
@@ -243,7 +189,6 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         LOG.debug("MidonetPluginV2.create_subnet exiting: subnet=%r", s)
         return s
 
-    @db_api.retry_if_session_inactive()
     def delete_subnet(self, context, id):
         LOG.debug("MidonetPluginV2.delete_subnet called: id=%s", id)
 
@@ -255,18 +200,11 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
 
         LOG.debug("MidonetPluginV2.delete_subnet exiting")
 
-    @db_api.retry_if_session_inactive()
     def update_subnet(self, context, id, subnet):
         LOG.debug("MidonetPluginV2.update_subnet called: id=%s", id)
 
         with context.session.begin(subtransactions=True):
             s = super(MidonetPluginV2, self).update_subnet(context, id, subnet)
-            self.extension_manager.process_update_subnet(context,
-                subnet['subnet'], s)
-            # NOTE(yamamoto): Retrieve the db object to get the correct
-            # revision
-            context.session.flush()
-            s = self.get_subnet(context, id)
             self.client.update_subnet_precommit(context, id, s)
 
         self.client.update_subnet_postcommit(id, s)
@@ -274,23 +212,13 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
         LOG.debug("MidonetPluginV2.update_subnet exiting: subnet=%r", s)
         return s
 
-    @db_api.retry_if_session_inactive()
     def create_port(self, context, port):
         LOG.debug("MidonetPluginV2.create_port called: port=%r", port)
 
         port_data = port['port']
-        # REVISIT(yamamoto): this nested transaction is a workaround
-        # for bug #1490917.
-        with db_api.autonested_transaction(context.session):
-            # Set status along admin_state_up if the parameter is specified.
-            if port['port'].get('admin_state_up') is not None:
-                if not port['port']['admin_state_up']:
-                    port['port']['status'] = n_const.PORT_STATUS_DOWN
+        with context.session.begin(subtransactions=True):
             # Create a Neutron port
-            port_db = self.create_port_db(context, port)
-            new_port = self._make_port_dict(port_db, process_extensions=False)
-            self.extension_manager.process_create_port(context, port_data,
-                                                       new_port)
+            new_port = super(MidonetPluginV2, self).create_port(context, port)
 
             # Do not create a gateway port if it has no IP address assigned as
             # MidoNet does not yet handle this case.
@@ -310,22 +238,8 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
             # Update fields
             port_data.update(new_port)
 
-            port_psec, has_ip = self._determine_port_security_and_has_ip(
-                context, port['port'])
-            port['port'][psec.PORTSECURITY] = port_psec
-            self._process_port_port_security_create(context,
-                                                    port['port'],
-                                                    new_port)
-
-            if port_psec is False:
-                if self._check_update_has_security_groups(port):
-                    raise psec.PortSecurityAndIPRequiredForSecurityGroups()
-                if self._check_update_has_allowed_address_pairs(port):
-                    raise addr_pair.AddressPairAndPortSecurityRequired()
-            else:
-                # Bind security groups to the port
-                self._ensure_default_security_group_on_port(context, port)
-
+            # Bind security groups to the port
+            self._ensure_default_security_group_on_port(context, port)
             sg_ids = self._get_security_groups_on_port(context, port)
             self._process_port_create_security_group(context, new_port, sg_ids)
 
@@ -359,28 +273,22 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
                     LOG.exception(_LE("Failed to delete port %s"),
                                   new_port['id'])
 
-        self._apply_dict_extend_functions('ports', new_port, port_db)
-
         LOG.debug("MidonetPluginV2.create_port exiting: port=%r", new_port)
         return new_port
 
-    @db_api.retry_if_session_inactive()
     def delete_port(self, context, id, l3_port_check=True):
         LOG.debug("MidonetPluginV2.delete_port called: id=%(id)s "
                   "l3_port_check=%(l3_port_check)r",
                   {'id': id, 'l3_port_check': l3_port_check})
 
-        l3plugin = directory.get_plugin(n_const.L3)
-
         # if needed, check to see if this is a port owned by
         # and l3-router.  If so, we should prevent deletion.
-        if l3_port_check and l3plugin:
-            l3plugin.prevent_l3_port_deletion(context, id)
+        if l3_port_check:
+            self.prevent_l3_port_deletion(context, id)
 
         with context.session.begin(subtransactions=True):
-            if l3plugin:
-                l3plugin.disassociate_floatingips(context, id,
-                                                  do_notify=False)
+            super(MidonetPluginV2, self).disassociate_floatingips(
+                context, id, do_notify=False)
             super(MidonetPluginV2, self).delete_port(context, id)
             self.client.delete_port_precommit(context, id)
 
@@ -388,7 +296,6 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
 
         LOG.debug("MidonetPluginV2.delete_port exiting: id=%r", id)
 
-    @db_api.retry_if_session_inactive()
     def update_port(self, context, id, port):
         LOG.debug("MidonetPluginV2.update_port called: id=%(id)s "
                   "port=%(port)r", {'id': id, 'port': port})
@@ -398,9 +305,6 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
             # update the port DB
             original_port = super(MidonetPluginV2, self).get_port(context, id)
             p = super(MidonetPluginV2, self).update_port(context, id, port)
-            c_utils.check_update_port(original_port, p)
-            self.extension_manager.process_update_port(context, port['port'],
-                                                       p)
 
             has_sg = self._check_update_has_security_groups(port)
             delete_sg = self._check_update_deletes_security_groups(port)
@@ -419,75 +323,350 @@ class MidonetPluginV2(plugin.MidonetMixinBase,
             self.update_address_pairs_on_port(context, id, port,
                                               original_port, p)
 
-            self._process_port_port_security_update(context, port['port'], p)
-
-            port_psec = p.get(psec.PORTSECURITY)
-            if port_psec is False:
-                if p.get(ext_sg.SECURITYGROUPS):
-                    raise psec.PortSecurityPortHasSecurityGroup()
-                if p.get(addr_pair.ADDRESS_PAIRS):
-                    raise addr_pair.AddressPairAndPortSecurityRequired()
-            # NOTE(yamamoto): Retrieve the db object to get the correct
-            # revision
-            context.session.flush()
-            context.session.expire_all()
-            p = self.get_port(context, id)
             self.client.update_port_precommit(context, id, p)
 
-        try:
-            self.client.update_port_postcommit(id, p)
-            if p['status'] != n_const.PORT_STATUS_NOTAPPLICABLE:
-                # Set status along admin_state_up to update.
-                if p['admin_state_up']:
-                    update_status = n_const.PORT_STATUS_ACTIVE
-                else:
-                    update_status = n_const.PORT_STATUS_DOWN
-
-                data = {'port': {'status': update_status}}
-                new_port = super(MidonetPluginV2, self).update_port(
-                    context, id, data)
-                # prevent binding:profile information from lacking
-                p['status'] = new_port['status']
-                if 'revision_number' in new_port:
-                    p['revision_number'] = new_port['revision_number']
-        except Exception as ex:
-            with excutils.save_and_reraise_exception():
-                LOG.error(_LE("Failed to update a port %(port_id)s in "
-                              "MidoNet: %(err)s"), {"port_id": id, "err": ex})
-                try:
-                    data = {'port': {'status': n_const.PORT_STATUS_ERROR}}
-                    super(MidonetPluginV2, self).update_port(context, id, data)
-                except Exception:
-                    LOG.exception(_LE("Failed to update port status %s"), id)
+        self.client.update_port_postcommit(id, p)
 
         LOG.debug("MidonetPluginV2.update_port exiting: p=%r", p)
         return p
 
-    @log_helpers.log_method_call
-    def create_security_group_rule_bulk(self, context, security_group_rules):
-        return super(MidonetPluginV2,
-                     self).create_security_group_rule_bulk_native(
-                     context, security_group_rules)
+    def create_router(self, context, router):
+        LOG.debug("MidonetPluginV2.create_router called: router=%(router)s",
+                  {"router": router})
 
-    def _midonet_v2_extend_network_dict(self, result, netdb):
-        session = db_api.get_session()
-        with session.begin(subtransactions=True):
-            self.extension_manager.extend_network_dict(session, netdb, result)
+        with context.session.begin(subtransactions=True):
+            r = super(MidonetPluginV2, self).create_router(context, router)
+            self.client.create_router_precommit(context, r)
 
-    def _midonet_v2_extend_port_dict(self, result, portdb):
-        session = db_api.get_session()
-        with session.begin(subtransactions=True):
-            self.extension_manager.extend_port_dict(session, portdb, result)
+        try:
+            self.client.create_router_postcommit(r)
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE("Failed to create a router %(r_id)s in Midonet:"
+                              "%(err)s"), {"r_id": r["id"], "err": ex})
+                try:
+                    self.delete_router(context, r['id'])
+                except Exception:
+                    LOG.exception(_LE("Failed to delete a router %s"), r["id"])
 
-    def _midonet_v2_extend_subnet_dict(self, result, subnetdb):
-        session = db_api.get_session()
-        with session.begin(subtransactions=True):
-            self.extension_manager.extend_subnet_dict(
-                session, subnetdb, result)
+        LOG.debug("MidonetPluginV2.create_router exiting: router=%(router)s.",
+                  {"router": r})
+        return r
 
-    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
-               attributes.NETWORKS, ['_midonet_v2_extend_network_dict'])
-    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
-               attributes.PORTS, ['_midonet_v2_extend_port_dict'])
-    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
-               attributes.SUBNETS, ['midonet_v2_extend_subnet_dict'])
+    def update_router(self, context, id, router):
+        LOG.debug("MidonetPluginV2.update_router called: id=%(id)s "
+                  "router=%(router)r", {"id": id, "router": router})
+
+        with context.session.begin(subtransactions=True):
+            r = super(MidonetPluginV2, self).update_router(context, id, router)
+            self.client.update_router_precommit(context, id, r)
+
+        self.client.update_router_postcommit(id, r)
+
+        LOG.debug("MidonetPluginV2.update_router exiting: router=%r", r)
+        return r
+
+    def delete_router(self, context, id):
+        LOG.debug("MidonetPluginV2.delete_router called: id=%s", id)
+
+        with context.session.begin(subtransactions=True):
+            super(MidonetPluginV2, self).delete_router(context, id)
+            self.client.delete_router_precommit(context, id)
+
+        self.client.delete_router_postcommit(id)
+
+        LOG.debug("MidonetPluginV2.delete_router exiting: id=%s", id)
+
+    def add_router_interface(self, context, router_id, interface_info):
+        LOG.debug("MidonetPluginV2.add_router_interface called: "
+                  "router_id=%(router_id)s, interface_info=%(interface_info)r",
+                  {'router_id': router_id, 'interface_info': interface_info})
+
+        with context.session.begin(subtransactions=True):
+            info = super(MidonetPluginV2, self).add_router_interface(
+                context, router_id, interface_info)
+            self.client.add_router_interface_precommit(context, router_id,
+                                                       info)
+
+        try:
+            self.client.add_router_interface_postcommit(router_id, info)
+        except Exception as ex:
+            LOG.error(_LE("Failed to create MidoNet resources to add router "
+                          "interface. info=%(info)s, router_id=%(router_id)s, "
+                          "error=%(err)r"),
+                      {"info": info, "router_id": router_id, "err": ex})
+            with excutils.save_and_reraise_exception():
+                self.remove_router_interface(context, router_id, info)
+
+        LOG.debug("MidonetPluginV2.add_router_interface exiting: info=%r",
+                  info)
+        return info
+
+    def remove_router_interface(self, context, router_id, interface_info):
+        LOG.debug("MidonetPluginV2.remove_router_interface called: "
+                  "router_id=%(router_id)s, interface_info=%(interface_info)r",
+                  {'router_id': router_id, 'interface_info': interface_info})
+
+        with context.session.begin(subtransactions=True):
+            info = super(MidonetPluginV2, self).remove_router_interface(
+                context, router_id, interface_info)
+            self.client.remove_router_interface_precommit(context, router_id,
+                                                          info)
+
+        self.client.remove_router_interface_postcommit(router_id, info)
+
+        LOG.debug("MidonetPluginV2.remove_router_interface exiting: info=%r",
+                  info)
+        return info
+
+    def create_floatingip(self, context, floatingip):
+        LOG.debug("MidonetPluginV2.create_floatingip called: ip=%r",
+                  floatingip)
+
+        with context.session.begin(subtransactions=True):
+            fip = super(MidonetPluginV2, self).create_floatingip(context,
+                                                              floatingip)
+            self.client.create_floatingip_precommit(context, fip)
+
+        try:
+            self.client.create_floatingip_postcommit(fip)
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE("Failed to create floating ip %(fip)s: %(err)s"),
+                          {"fip": fip, "err": ex})
+                try:
+                    self.delete_floatingip(context, fip['id'])
+                except Exception:
+                    LOG.exception(_LE("Failed to delete a floating ip %s"),
+                                  fip['id'])
+
+        LOG.debug("MidonetPluginV2.create_floatingip exiting: fip=%r", fip)
+        return fip
+
+    def delete_floatingip(self, context, id):
+        LOG.debug("MidonetPluginV2.delete_floatingip called: id=%s", id)
+
+        with context.session.begin(subtransactions=True):
+            super(MidonetPluginV2, self).delete_floatingip(context, id)
+            self.client.delete_floatingip_precommit(context, id)
+
+        self.client.delete_floatingip_postcommit(id)
+
+        LOG.debug("MidonetPluginV2.delete_floatingip exiting: id=%r", id)
+
+    def update_floatingip(self, context, id, floatingip):
+        LOG.debug("MidonetPluginV2.update_floatingip called: id=%(id)s "
+                  "floatingip=%(floatingip)s ",
+                  {'id': id, 'floatingip': floatingip})
+
+        with context.session.begin(subtransactions=True):
+            fip = super(MidonetPluginV2, self).update_floatingip(context, id,
+                                                              floatingip)
+            self.client.update_floatingip_precommit(context, id, fip)
+
+            # Update status based on association
+            if fip.get('port_id') is None:
+                fip['status'] = n_const.FLOATINGIP_STATUS_DOWN
+            else:
+                fip['status'] = n_const.FLOATINGIP_STATUS_ACTIVE
+            self.update_floatingip_status(context, id, fip['status'])
+
+        self.client.update_floatingip_postcommit(id, fip)
+
+        LOG.debug("MidonetPluginV2.update_floating_ip exiting: fip=%s", fip)
+        return fip
+
+    def create_security_group(self, context, security_group, default_sg=False):
+        LOG.debug("MidonetPluginV2.create_security_group called: "
+                  "security_group=%(security_group)s "
+                  "default_sg=%(default_sg)s ",
+                  {'security_group': security_group, 'default_sg': default_sg})
+
+        sg = security_group.get('security_group')
+        tenant_id = self._get_tenant_id_for_create(context, sg)
+        if not default_sg:
+            self._ensure_default_security_group(context, tenant_id)
+
+        # Create the Neutron sg first
+        with context.session.begin(subtransactions=True):
+            sg = super(MidonetPluginV2, self).create_security_group(
+                context, security_group, default_sg)
+            self.client.create_security_group_precommit(context, sg)
+
+        try:
+            self.client.create_security_group_postcommit(sg)
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE("Failed to create MidoNet resources for "
+                              "sg %(sg)r, error=%(err)r"),
+                          {"sg": sg, "err": ex})
+                try:
+                    self.delete_security_group(context, sg['id'])
+                except Exception:
+                    LOG.exception(_LE("Failed to delete a security group %s"),
+                                  sg['id'])
+
+        LOG.debug("MidonetPluginV2.create_security_group exiting: sg=%r", sg)
+        return sg
+
+    def delete_security_group(self, context, id):
+        LOG.debug("MidonetPluginV2.delete_security_group called: id=%s", id)
+
+        sg = super(MidonetPluginV2, self).get_security_group(context, id)
+        if not sg:
+            raise ext_sg.SecurityGroupNotFound(id=id)
+
+        if sg["name"] == 'default' and not context.is_admin:
+            raise ext_sg.SecurityGroupCannotRemoveDefault()
+
+        with context.session.begin(subtransactions=True):
+            super(MidonetPluginV2, self).delete_security_group(context, id)
+            self.client.delete_security_group_precommit(context, id)
+
+        self.client.delete_security_group_postcommit(id)
+
+        LOG.debug("MidonetPluginV2.delete_security_group exiting: id=%r", id)
+
+    def create_security_group_rule(self, context, security_group_rule):
+        LOG.debug("MidonetPluginV2.create_security_group_rule called: "
+                  "security_group_rule=%(security_group_rule)r",
+                  {'security_group_rule': security_group_rule})
+
+        with context.session.begin(subtransactions=True):
+            rule = super(MidonetPluginV2, self).create_security_group_rule(
+                context, security_group_rule)
+            self.client.create_security_group_rule_precommit(context, rule)
+
+        try:
+            self.client.create_security_group_rule_postcommit(rule)
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE('Failed to create security group rule %(sg)s,'
+                          'error: %(err)s'), {'sg': rule, 'err': ex})
+                try:
+                    self.delete_security_group_rule(context, rule['id'])
+                except Exception:
+                    LOG.exception(_LE("Failed to delete "
+                                      "a security group rule %s"), rule['id'])
+
+        LOG.debug("MidonetPluginV2.create_security_group_rule exiting: "
+                  "rule=%r", rule)
+        return rule
+
+    def create_security_group_rule_bulk(self, context, rules):
+        LOG.debug("MidonetPluginV2.create_security_group_rule_bulk called: "
+                  "security_group_rules=%(security_group_rules)r",
+                  {'security_group_rules': rules})
+
+        with context.session.begin(subtransactions=True):
+            rules = super(
+                MidonetPluginV2, self).create_security_group_rule_bulk_native(
+                    context, rules)
+            self.client.create_security_group_rule_bulk_precommit(context,
+                                                                  rules)
+
+        try:
+            self.client.create_security_group_rule_bulk_postcommit(rules)
+        except Exception as ex:
+            LOG.error(_LE("Failed to create bulk security group rules %(sg)s, "
+                          "error: %(err)s"), {"sg": rules, "err": ex})
+            with excutils.save_and_reraise_exception():
+                for rule in rules:
+                    self.delete_security_group_rule(context, rule['id'])
+
+        LOG.debug("MidonetPluginV2.create_security_group_rule_bulk exiting: "
+                  "rules=%r", rules)
+        return rules
+
+    def delete_security_group_rule(self, context, sg_rule_id):
+        LOG.debug("MidonetPluginV2.delete_security_group_rule called: "
+                  "sg_rule_id=%s", sg_rule_id)
+
+        with context.session.begin(subtransactions=True):
+            super(MidonetPluginV2, self).delete_security_group_rule(context,
+                                                                 sg_rule_id)
+            self.client.delete_security_group_rule_precommit(context,
+                                                             sg_rule_id)
+
+        self.client.delete_security_group_rule_postcommit(sg_rule_id)
+
+        LOG.debug("MidonetPluginV2.delete_security_group_rule exiting: id=%r",
+                  sg_rule_id)
+
+    def create_agent_membership(self, context, agent_membership):
+        LOG.debug("MidonetPluginV2.create_agent_membership called: "
+                  " %(agent_membership)r",
+                  {'agent_membership': agent_membership})
+
+        with context.session.begin(subtransactions=True):
+            am = super(MidonetPluginV2, self).create_agent_membership(
+                context, agent_membership)
+            self.client.create_agent_membership_precommit(context, am)
+
+        try:
+            self.client.create_agent_membership_postcommit(am)
+        except Exception as ex:
+            with excutils.save_and_reraise_exception():
+                LOG.error(_LE("Failed to create agent membership. am: %(am)r, "
+                              "error: %(err)s"), {'am': am, 'err': ex})
+                try:
+                    self.delete_agent_membership(context, am['id'])
+                except Exception:
+                    LOG.exception(_LE("Failed to delete "
+                                      "an agent membership %s"), am['id'])
+
+        LOG.debug("MidonetPluginV2.create_agent_membership exiting: "
+                  "%(agent_membership)r", {'agent_membership': am})
+        return am
+
+    def get_agent_membership(self, context, id, filters=None, fields=None):
+        LOG.debug("MidonetPluginV2.get_agent_membership called: id=%(id)r",
+                  {'id': id})
+
+        with context.session.begin(subtransactions=True):
+            am = super(MidonetPluginV2, self).get_agent_membership(context, id)
+
+        LOG.debug("MidonetPluginV2.get_agent_membership exiting: id=%(id)r, "
+                  "agent_membership=%(agent_membership)r",
+                  {'id': id, 'agent_membership': am})
+        return am
+
+    def get_agent_memberships(self, context, filters=None, fields=None,
+                              sorts=None, limit=None, marker=None,
+                              page_reverse=False):
+        LOG.debug("MidonetPluginV2.get_agent_memberships called")
+
+        with context.session.begin(subtransactions=True):
+            ams = super(MidonetPluginV2, self).get_agent_memberships(
+                context, filters, fields, sorts, limit, marker, page_reverse)
+
+        LOG.debug("MidonetPluginV2.get_agent_memberships exiting")
+        return ams
+
+    def delete_agent_membership(self, context, id):
+        LOG.debug("MidonetPluginV2.delete_agent_membership called: %(id)r",
+                  {'id': id})
+
+        with context.session.begin(subtransactions=True):
+            super(MidonetPluginV2, self).delete_agent_membership(context, id)
+            self.client.delete_agent_membership_precommit(context, id)
+
+        self.client.delete_agent_membership_postcommit(id)
+
+        LOG.debug("MidonetPluginV2.delete_agent_membership exiting: %(id)r",
+                  {'id': id})
+
+    def get_agents(self, context, filters=None, fields=None):
+        LOG.debug("MidonetPluginV2.get_agents called")
+
+        agents = super(MidonetPluginV2, self).get_agents(context, filters,
+                                                         fields)
+        return agents + self.client.get_agents()
+
+    def get_agent(self, context, id, fields=None):
+        LOG.debug("MidonetPluginV2.get_agent called: %(id)r", {'id': id})
+
+        agent = self.client.get_agent(id)
+        if not agent:
+            agent = super(MidonetPluginV2, self).get_agent(context, id, fields)
+        return agent
